/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.collections.AST;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ReferenceAttributeType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.ScopeException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;

import java.util.*;

/**
 * Semantische Überprüfung von Attributdefinitionen von Realen Attributen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RealAttributeSemantics implements SemanticChecks {

	private AttributeSemantics _attributeSemantics;

	private Semantics _semantics;

	private MemberCheck _check;

	private ScopeSemantics _scopeSemantics;

	private ProtocolType _protocolType = RealAttributeSemantics.ProtocolType.UNKOWN;

	/** Typ oder Objekttyp for den das Protokoll erstellt wird */
	private SystemObject _protocolObject;

	private enum ProtocolType {
		UNKOWN("Unbekannt"),
		/** Protokoll mit Pseudoobjekten */
		PSEUDOOBJECT("Pseudo-Objekt"),
		/** Protokoll mit einem festen Objekt */
		OBJECT("Objekt");

		private final String _readableDescription;

		ProtocolType(String readableDescription) {
			_readableDescription = readableDescription;
		}

		public String getReadableDescription() {
			return _readableDescription;
		}
	}

	/**
	 * @param semantics
	 * @param attributeSemantics
	 * @param scopeSemantics
	 */
	public RealAttributeSemantics(Semantics semantics, AttributeSemantics attributeSemantics, ScopeSemantics scopeSemantics) {
		this._semantics = semantics;
		this._check = semantics.getCheck();
		this._attributeSemantics = attributeSemantics;
		this._scopeSemantics = scopeSemantics;
	}

	/**
	 * Fügt der {@link Semantics#pi} ein (Listen-)Attribut aus dem Spalten-Bereich hinzu. Überprüft ob das Attribut korrekt qualifiziert wurde Es wird überprüft,
	 * ob das Attribut voll qualifiziert wird bzw. ob eine Qualifizierung unter zu Hilfenahme der auf dem Stack liegenden Qualifizierung möglich ist.
	 *
	 * @param ast        Knoten (wird nur für Zeilenangabe in der Fehlermeldung benötigt)
	 * @param parts      Qualifizierung des Attributs
	 * @param columnName Name der dem Atribut im Skript mittels "als" zugewiesen wird
	 * @param padData    <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen.
	 *
	 * @throws CriticalParserException Falls es zu einem Kommuniktaionsfehler kommt oder das Attribut nicht korrekt qualifiziert ist
	 */
	public void addRealAttribute(AST ast, List<String> parts, String columnName, boolean padData) throws CriticalParserException {
		try {
			final RealAttributeDescription[] attributeDesc;
			attributeDesc = _attributeSemantics.storeAttributeInformation(createRealAttributeDescription(parts, padData));

			for(RealAttributeDescription anAttributeDesc : attributeDesc) {
				anAttributeDesc.setColumnName(columnName);
				_semantics.pi.addElement(anAttributeDesc);
				// Zähler hochsetzen: Reales Attribut wird verwendet
				_attributeSemantics.incAttributeCount();
			}
		}
		catch(CriticalParserException criticalException) {
			criticalException.line = ast.getLine();
			throw criticalException;
		}
	}

	/**
	 * Baut die Qualifizierung für ein Attribut auf.
	 *
	 * @param type    Typ des ersten Elements von parts
	 * @param quali   Qualifizierung die verwendet werden soll, um die fehlende Qualifizierung in parts auszugleichen
	 * @param parts   Qualifizierung des Attributs aus dem Skript
	 * @param padData <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen.
	 *
	 * @return Attributbeschreibung mit vollständiger Qualifizierung, die jedoch noch nicht auf Semantik geprüft wurde!
	 *
	 * @throws CriticalParserException Falls es in der Qualifizierung eine ungültige Zeichenkette gefunden wird
	 */
	private RealAttributeDescription buildAttributeDescription(int type, Qualifier quali, List<String> parts, boolean padData) throws CriticalParserException {
		RealAttributeDescription attributeDesc = new RealAttributeDescription(_semantics.pi.getSimVar(), padData);
		int index = 0; //begin der Referenz in parts
		int guessType = parts.size() % 3;

		if(quali == null && !(guessType == 1 && (type == MemberCheck.OBJECT || type
		                                                                       == MemberCheck.OBJECTTYPE)))  //nur für den Fall, dass eine vollständige Qualifizierung angegeben ist, darf quali null sein.
		{
			throw new ScopeException(InterpreterErrorMessage.OBJECT_OR_OBJECTTYPE_EXPECTED + parts.get(0));
		}
		//		else
		{
			//type ist der Typ des ersten Elements in parts
			//je nachdem, von welchem Typ das erste Element in parts ist, muss ein anderer Typ aus der Qualifizierung vom Stack geholt werden
			//z.B. erster Typ in Parts ist ein Apsekt. Dann muss in 'quali' die Attributgruppe und das Objekt dieses Aspekts stehen.
			switch(type) {
				case MemberCheck.OBJECT:
				case MemberCheck.OBJECTTYPE:
					if(guessType == 0)				//obj:asp:att
					{
						if(parts.size() >= 3) {
							attributeDesc.qualifier.object = quali.object;
							attributeDesc.qualifier.attributeGroup = quali.attributeGroup;
							if(_check.whatIs(parts.get(1)) == MemberCheck.FREE_ASPECT) {
								attributeDesc.qualifier.freeAspect = parts.get(1);
							}
							else {
								attributeDesc.qualifier.aspect = _check.getAspect(parts.get(1));
							}
							attributeDesc.setAttributeString(parts.get(2), attributeDesc.qualifier.attributeGroup);
							index = 3;
						}
						else {
							throw new ScopeException(InterpreterErrorMessage.INCOMPLETE_SCOPE);
						}
					}
					else if(guessType == 1)		//obj:atg:asp:att
					{
						if(parts.size() >= 4) {
							attributeDesc.qualifier.object = _check.getObject(parts.get(0));
							attributeDesc.qualifier.attributeGroup = _check.getAttributeGroup(parts.get(1));
							if(_check.whatIs(parts.get(2)) == MemberCheck.FREE_ASPECT) {
								attributeDesc.qualifier.freeAspect = parts.get(2);
							}
							else {
								attributeDesc.qualifier.aspect = _check.getAspect(parts.get(2));
							}

							attributeDesc.setAttributeString(parts.get(3), attributeDesc.qualifier.attributeGroup);
							index = 4;
						}
						else {
							throw new ScopeException(InterpreterErrorMessage.INCOMPLETE_SCOPE);
						}
					}
					else							//obj:att
					{
						if(parts.size() >= 2) {
							attributeDesc.qualifier.object = quali.object;
							attributeDesc.qualifier.attributeGroup = quali.attributeGroup;
							attributeDesc.qualifier.aspect = quali.aspect;
							attributeDesc.qualifier.freeAspect = quali.freeAspect;
							attributeDesc.setAttributeString(parts.get(1), attributeDesc.qualifier.attributeGroup);
							index = 2;
						}
						else {
							throw new ScopeException(InterpreterErrorMessage.INCOMPLETE_SCOPE);
						}
					}
					break;

				case MemberCheck.ATTRIBUTEGROUP:							//atg:asp:att
					if(parts.size() >= 3) {
						attributeDesc.qualifier.object = quali.object;
						attributeDesc.qualifier.attributeGroup = _check.getAttributeGroup(parts.get(0));
						if(_check.whatIs(parts.get(1)) == MemberCheck.FREE_ASPECT) {
							attributeDesc.qualifier.freeAspect = parts.get(1);
						}
						else {
							attributeDesc.qualifier.aspect = _check.getAspect(parts.get(1));
						}

						attributeDesc.setAttributeString(parts.get(2), attributeDesc.qualifier.attributeGroup);
						index = 3;
					}
					else {
						throw new ScopeException(InterpreterErrorMessage.INCOMPLETE_SCOPE);
					}
					break;

				case MemberCheck.ASPECT:
					if(parts.size() >= 2)		//asp:att
					{
						attributeDesc.qualifier.object = quali.object;
						attributeDesc.qualifier.attributeGroup = quali.attributeGroup;
						attributeDesc.qualifier.aspect = _check.getAspect(parts.get(0));
						attributeDesc.setAttributeString(parts.get(1), attributeDesc.qualifier.attributeGroup);
						index = 2;
					}
					else {
						throw new ScopeException(InterpreterErrorMessage.INCOMPLETE_SCOPE);
					}
					break;
				case MemberCheck.FREE_ASPECT:								//asp:att
					if(parts.size() >= 2) {
						attributeDesc.qualifier.object = quali.object;
						attributeDesc.qualifier.attributeGroup = quali.attributeGroup;
						attributeDesc.qualifier.freeAspect = parts.get(0);
						attributeDesc.setAttributeString(parts.get(1), attributeDesc.qualifier.attributeGroup);
						index = 2;
					}
					else {
						throw new ScopeException(InterpreterErrorMessage.INCOMPLETE_SCOPE);
					}
					break;
				case MemberCheck.ATTRIBUTE:									//att
					attributeDesc.qualifier.object = quali.object;
					attributeDesc.qualifier.attributeGroup = quali.attributeGroup;
					attributeDesc.qualifier.aspect = quali.aspect;
					attributeDesc.qualifier.freeAspect = quali.freeAspect;
					attributeDesc.setAttributeString(parts.get(0), attributeDesc.qualifier.attributeGroup);
					index = 1;
					break;
				default: //sollte niemals aufgerufen werden, aber wer weiss das schon so genau ;)
					throw new ScopeException(InterpreterErrorMessage.UNKNOWN_TYPE + type);
			}
		}


		// prüfen, ob es sich um eine Referenz handelt:
		RealAttributeDescription base = attributeDesc;

		while(index != parts.size()) {
			// es gibt noch weitere Einträge in der Liste

			if(base.getAttribute() == null) {
				throw new ScopeException(InterpreterErrorMessage.INVALID_SCOPE + base.qualifier.attributeGroup.getPid() + " : " + base.getName());
			}
			else {
				if(base.getAttribute().getAttributeType() instanceof ReferenceAttributeType) {
					if((parts.size() - index) % 3 == 0) {
						RealAttributeDescription reference = new RealAttributeDescription(_semantics.pi.getSimVar(), padData);

						//Obj, atg, asp & att eintragen
						final SystemObjectType referencedObjectType = ((ReferenceAttributeType)base.getAttribute().getAttributeType()).getReferencedObjectType();
						final AttributeGroup attributeGroup = _semantics.model.getAttributeGroup(parts.get(index));

						
						if(!referencedObjectType.getAttributeGroups().contains(attributeGroup)) {
							throw new SemanticErrorException(InterpreterErrorMessage.MISMATCHED_ATTRIBUTEGROUP + attributeGroup);
						}

						reference.qualifier.object = referencedObjectType;
						reference.qualifier.attributeGroup = attributeGroup;

						if(_check.whatIs(parts.get(index + 1)) == MemberCheck.FREE_ASPECT) {
							reference.qualifier.freeAspect = parts.get(index + 1);
						}
						else {
							reference.qualifier.aspect = _semantics.model.getAspect(parts.get(index + 1));
						}

						reference.setAttributeString(parts.get(index + 2), reference.qualifier.attributeGroup);

						base.setReference(reference);

						index += 3;
						base = reference;
					}
					else {
						throw new ScopeException(InterpreterErrorMessage.INVALID_REFERENCE + parts.subList(index, parts.size()));
					}
				}
				else	//Keine Referenz, trotzdem noch weitere Elemente in der Liste
				{
					throw new SemanticErrorException(InterpreterErrorMessage.REFERENCE_EXPECTED);
				}
			}
		}


		return attributeDesc;
	}

	/**
	 * Erzeugt Beschreibungen von Realen Attributen. Dabei wird die Qualifizierung der Attribute überprüft. Wird ein Alias verwendet, so wird er aufgelöst. Daher
	 * ist es moeglich, dass meherer Attribute zurueckgegeben werden: Wenn einem Alias mittels Pseudoobjekt-Definition mehrere Objekte zugewiesen wurden. (=> Wird
	 * ein Alias verwendet, werden alle durch die Pseudoobjektdefinitionen erzeugten Attribute zurückgegeben.)
	 *
	 * @param parts   Die Qualifizierung des Attributs. Die Liste wird durch diese Methode nicht geändert!
	 * @param padData <code>true</code> falls Daten durch die Datenaufbereitung aufgefüllt werden sollen.
	 *
	 * @return Array mit den gebildeten Attributbeschreibungen
	 *
	 * @throws CriticalParserException Falls es zu einem Fehler kam (Kommuniktation, fehlerhafte Qualifizierung)
	 */
	RealAttributeDescription[] createRealAttributeDescription(final List<String> parts, final boolean padData) throws CriticalParserException {

		RealAttributeDescription attributeDesc = null;
		final int type;
		final Qualifier quali;

		/*
				 * Folgende Qualifizierungen sind möglich
				 * parts.size():	Pfad
				 * 1:	Attribut				(benutze Qualifizierung aus Stack)
				 * 2:	Asp:Att					(benutze Qualifizierung aus Stack)
				 * 3:	Atg:Asp:Att				(benutze Qualifizierung aus Stack)
				 * 4:	Obj:Atg:Asp:Att
				 *
				 * sowie bei Referenzen: 		(das dritte oder viertletzte Element muss ein Objekttyp sein!)
				 * 5:	Att:Obj:Att:Asp:Att 	(benutze Qualifizierung aus Stack)
				 * 6:	...
				 *
				 * sowie durch TAnf S. 19 gefordert:
				 * 7: Obj:Att				    (benutze Qualifizierung aus dem Stack, die besagtes Objekt verwendet)
				 * 8: Obj:Asp:Att
				 */

		// Überprüfen ob ein Alias verwendet wird

		// Der Text der im ersten Segment steht (kann alias sein oder nicht)
		final String node = parts.get(0);

		// aliasType ist null falls kein Alisas definiert wurde (es sich also um ein normales Objekt bzw. Objekttyp handelt). Ansonsten enthält es den Objekttyp
		final String aliasType = _semantics.symbolTable.getAlias(node);

		final String[] objects;

		if(aliasType != null) {
			//Es wird ein Alias verwendet
			//Alias durch Objekttyp austauschen
			parts.set(0, aliasType);

			// Die zugehörigen Pseudo-Objekte holen
			objects = _semantics.symbolTable.getAliasObjects(node);
		}
		else {
			objects = null;
		}

		/*
				 * Darum ist der folgende Code so umständlich:
				 * Einfachster Weg: Elemente aus parts auslesen, und mittels DataModel.getObject(pid) rausbekommen, von welchem Typ sie sind.
				 * Pro: Simpel.
				 * Contra: Langsam, da wenn getObject(...) der Name eines Attributs übergeben wird, eine Warnung von der Konfiguration ausgegeben wird -> verlangsamt den Programmlauf -> Semantiküberprüfung mit 1000 Attributen dauert ca. 2 Sekunden!
				 *
				 * Daher: Länge des Eingabestrings lesen, und so herausbekommen, wo welches Element stehen sollte. -> Semantiküberprüfung mit 2000 Attributen (190kb) dauert etwa 1s; 2/3 der Zeit wird für das Parsen verwendet.
				 */

		//Länge prüfen:
		final int length = parts.size();
		final int mod = length % 3;				// % 3 da durch eine Referenz 3 Elementen hinzukommen: atg2, asp2 & att2

		if(mod == 0 || mod == 2)			// mod == 0 (obj:asp:att oder atg:asp:att) oder mod == 2 ( obj:atg oder asp:att)
		{
			//erstes Element überprüfen
			type = _check.whatIs(parts.get(0));
			if(type == MemberCheck.OBJECT || type == MemberCheck.OBJECTTYPE) {
				quali = _scopeSemantics._objectToScope.get(parts.get(0));
				if(quali == null) throw new ScopeException(InterpreterErrorMessage.INVALID_SCOPE + parts);
			}
			else {
				quali = _scopeSemantics.peek();
			}
			attributeDesc = buildAttributeDescription(type, quali, parts, padData);
		}
		else	// obj:atg:asp:att oder att
		{
			//erstes Element kann nicht mit whatIs() überprüft werden, da es sich ja um ein Attribut handeln könnte
			if(length >= 4) {
				//das erste Element muss überprüft werden, sonst ist keine Typ-Bestimmung möglich
				type = _check.whatIs(parts.get(0));
				quali = _scopeSemantics.peek();
				attributeDesc = buildAttributeDescription(type, quali, parts, padData);
			}
			else {
				//wenn die Länge 1 ist, sollte es sich um ein Attribut handeln
				quali = _scopeSemantics.peek();

				//falls quali null sein sollte, dann muss es sich doch um ein Objekt handeln (sonst Fehler)
				if(quali == null) {
					type = _check.whatIs(parts.get(0));
				}
				else {
					type = MemberCheck.ATTRIBUTE;	//Annahme: Es ist ein Attribut
				}

				attributeDesc = buildAttributeDescription(type, quali, parts, padData);
			}
		}

		//Qualifizierung prüfen
		attributeDesc.validate();

		checkAdd(aliasType != null, attributeDesc.qualifier.object);

		final RealAttributeDescription[] result;

		//falls ein Alias verwendet wurde, jetzt die jeweiligen Attribute bilden. Da dem Alias in der Pseudoobjekt-Definition nur Objekte vom gerade überprüften Typ zugeordnet werden können, stimmt deren Qualifizierung automatisch 
		if(objects != null) {
			result = new RealAttributeDescription[objects.length];
			for(int i = 0; i < objects.length; i++) {

				//Alias-Objekt gegen konkretes Objekt austauschen
				final String objectPid = objects[i];
				if(objectPid != null && !objectPid.isEmpty()){
					final SystemObject object = _check.getObject(objectPid);
					if(object == null) throw new SemanticErrorException(InterpreterErrorMessage.INVALID_OBJECT_PID + objectPid);
					if(!object.isOfType(aliasType)){
						throw new SemanticErrorException("Der Alias-Typ \"" + aliasType + "\" des Aliases \"" + node + "\" konnte nicht durch das Objekt \"" + object + "\" vom Typ \"" + object.getType() + "\" ersetzt werden.");
					}
					final RealAttributeDescription temp = attributeDesc.clone();
					temp.setIndex(i, node);
					temp.qualifier.object = object;
					result[i] = temp;
				}else{
					final RealAttributeDescription temp = attributeDesc.clone();
					temp.qualifier.object = null;
					temp.setIndex(i, node);
					result[i] = temp;
				}
			}
		}
		else {
			//kein Alias -> nur eine Attributbeschreibung

			//	Objekt gegen konkretes Objekt austauschen
			final SystemObject temp = _check.getObject(attributeDesc.qualifier.object);

			// falls keine Bindung existiert, dann Fehler auswerfen, falls das Protokoll ausgewertet werden soll!
			if(temp instanceof SystemObjectType && _semantics.processScript) {
				SystemObject mainObject = _semantics.getSymbolTable().getMainObject();
				if(mainObject == null){
					throw new SemanticErrorException("Es wurde kein Hauptobjekt übergeben, um dem Typ \"" + temp + "\" aufzulösen.");
				}
				throw new SemanticErrorException("Der Typ \"" + temp + "\" konnte nicht durch das Hauptobjekt \"" + mainObject + "\" vom Typ \"" + mainObject.getType() + "\" ersetzt werden.");
			}

			attributeDesc.qualifier.object = temp;

			result = new RealAttributeDescription[1];
			result[0] = attributeDesc;
		}
		return result;
	}

	void checkAdd(final boolean hasAlias, final SystemObject object) throws CriticalParserException {
		if(_protocolType == ProtocolType.UNKOWN) {
			if(hasAlias) {
				_protocolType = RealAttributeSemantics.ProtocolType.PSEUDOOBJECT;
			}
			else {
				_protocolType = RealAttributeSemantics.ProtocolType.OBJECT;
				_protocolObject = object;
			}
		}
		else {
			if(hasAlias) {
				if(_protocolType != RealAttributeSemantics.ProtocolType.PSEUDOOBJECT) {
					failIncompatibleColumnType(ProtocolType.PSEUDOOBJECT);
				}
			}
			else {
				if(_protocolType != RealAttributeSemantics.ProtocolType.OBJECT) {
					failIncompatibleColumnType(ProtocolType.OBJECT);
				}
				else if(!_protocolObject.equals(object)) {
					failMultipleObjectTypes(object);
				}
			}
		}
	}

	private void failIncompatibleColumnType(final ProtocolType incompatibleType1) throws CriticalParserException {
		throw new SemanticErrorException(
				InterpreterErrorMessage.INCOMPATIBLE_COLUMN_TYPE + "In dem Skript befindet sich eine " + _protocolType.getReadableDescription() + "- und eine "
				+ incompatibleType1.getReadableDescription() + "-Definition."
		);
	}


	private void failMultipleObjectTypes(final SystemObject object) throws CriticalParserException {
		throw new SemanticErrorException(
				InterpreterErrorMessage.MULTIPLE_OBJECTS + "Das Skript enthält sowohl Referenzen zu " + type(object) + " als auch zu " + type(_protocolObject) + "."
		);
	}

	private String type(final SystemObject object) {
		if(object instanceof SystemObjectType) {
			return object.toString();
		}
		return object + " vom Typ " + object.getType();
	}

	/**
	 * Tut nichts.
	 *
	 * @see de.bsvrz.pua.prot.interpreter.semantics.SemanticChecks#cleanUp(de.bsvrz.pua.prot.util.ProcessingParameter, boolean)
	 */
	public void cleanUp(ProcessingParameter pp, boolean processScript) throws SemanticErrorException {
	}
}
