/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing;

import de.bsvrz.pua.prot.client.dataobject.PuADataState;
import de.bsvrz.pua.prot.processing.util.*;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ExpressionResult.ResultType;
import de.bsvrz.pua.prot.util.ExpressionTree;
import de.bsvrz.pua.prot.util.ProcessingInformation.ApplyAggregation;
import de.bsvrz.pua.prot.util.ResultValue;
import de.bsvrz.pua.prot.util.attributes.Elements;
import de.bsvrz.pua.prot.util.attributes.RealElement;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

/**
 * InterMediateDataSetBuilder Erzeugt Ausgangsdatensätze und legt sie in einem Puffer ab.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class IMDSBuilder {

	/** Wert eines leeren Ausgangsdatensatzes. Wird verwendet, wenn ein leerer Datensatz erzeugt wird. */
	private static final ResultValue RV_EMPTY_DATASET = new ResultValue(new ExpressionResult("LEERER DATENSATZ"));

	/** Wert eines Datensatzes, der gefiltert wurde und daher nicht existiert */
	private static final ResultValue RV_FILTERED_DATASET = new ResultValue(new ExpressionResult());

	/** Wert eines wiederholten Ausgangsdatensatzes. Wird verwendet, wenn sich die Werte von mehreren Ausgangsdatensätzen nicht unterscheiden. */
	private static final ResultValue RV_REPETITION = new ResultValue(new ExpressionResult("WIEDERHOLUNG"));

	/** Debug-Ausgabe. */
	private static final Debug _debug = Debug.getLogger();

	/**
	 * Letzter "roher" Datensatz, also nicht aggregierte, nicht nachgefilterte Daten die durch den Filter gekommen sind. Darf nur in {@link
	 * #createIntermediateDataSet(boolean)} verwendet werden.
	 */
	private List<BaseDataSet> _lastBaseData = null;

	/** Puffer, der die Zuordnung: Reale Elemente -> Werte enthält. */
	protected RingBuffer<ValueContainer> _realElementValues;

	/** Objekt zur Statusberechnung. */
	private final Status _status;

	/** Führt die Datenaufbereitung durch. */
	private final ProcessingBuffer _processingBuffer;

	/** Im Spalten-Bereich definierte Attribute. */
	private final List<Elements> _attributes;

	/** Zuordnung der temporären Attribute zu ihren Wertebeschaffern. */
	private final Hashtable<TempAttributeDescription, TempAttribut> _tempElements;

	/** Zuordnung der realen Attribute zu ihren Wertebeschaffern. */
	private final Hashtable<RealElement, ValueProvider> _realElements;

	/** Im Skript definierte Einschränlung. */
	private final Tuple<ExpressionTree, Boolean> _constraint;

	/** Index des aktuellen Intervalls. */
	private int _currentInterval = 0;

	/** Aktueller Status. Wird von {@link #_status} berechnet. */
	private byte _currentStatus;

	/** <code>true</code> falls Datensätze verschmolzen werden sollen. */
	private final boolean _merge;

	/** true wenn NoChange-Informationen für Zustandsprotokolle pro Zelle generiert werden sollen (statt pro Zeile) */
	private final boolean _noChangeMarkerPerCell;

	/** Zeitpunkt Intervallende des zuletzt beendeten Intervalls. */
	private long _endOfInterval;


	/**
	 * Periodenindex, bei der zugehörige Buffer beginnt zu laden. Wird bei Datensatzerstellen auf _currentInterval addiert damit die Indized bei Onlinedaten nicht
	 * wieder bei 0 beginnen.
	 */
	private final int _startPeriodIndex;

	/**
	 * Initialisierung.
	 *  @param merge           <code>true</code> falls Datensätze verschmolzen werden sollen.
	 * @param noChangeMarkerPerCell  true wenn NoChange-Informationen für Zustandsprotokolle pro Zelle generiert werden sollen (statt pro Zeile)
	 * @param attributes       Im Spalten-Bereich definierte Attribute.
	 * @param constraint       Im Skript definierte Einschränlung.
	 * @param realElements     Zuordnung der realen Attribute zu ihren Wertebeschaffern.
	 * @param tempElements     Zuordnung der temporären Attribute zu ihren Wertebeschaffern.
	 * @param processingBuffer Datenaufbereiter.
	 * @param status           Objekt zur Statusberechung.
	 * @param startPeriodIndex
	 */
	public IMDSBuilder(
			final boolean merge,
			final boolean noChangeMarkerPerCell,
			final List<Elements> attributes,
			final Tuple<ExpressionTree, Boolean> constraint,
			final Hashtable<RealElement, ValueProvider> realElements,
			final Hashtable<TempAttributeDescription, TempAttribut> tempElements,
			final ProcessingBuffer processingBuffer,
			final Status status,
			final int startPeriodIndex) {
		this._merge = merge;
		_noChangeMarkerPerCell = noChangeMarkerPerCell;
		this._attributes = attributes;
		this._constraint = constraint;
		this._realElements = realElements;
		this._tempElements = tempElements;
		this._processingBuffer = processingBuffer;
		this._status = status;
		_startPeriodIndex = startPeriodIndex;
		_realElementValues = new RingBuffer<ValueContainer>(ProcessingBuffer.INITIAL_RINGBUFFER_SIZE, RingBuffer.UNBOUNDED_SIZE);
	}

	/**
	 * Sammelt die Werte der realen Attribute. Stellt dann einen Basisdatensatz mit den Werten der realen Attribute zusammen.
	 *
	 * @param insertEmpty     <code>true</code> legt fest, dass ein leerer Datensatz eingefügt werden soll.
	 * @param winnerTimeStamp Minimaler, nicht-aufgefüllter Zeitstempel.
	 */
	public void collectData(final boolean insertEmpty, final long winnerTimeStamp) {
		final ValueContainer vc = new ValueContainer(winnerTimeStamp, insertEmpty, _realElements.size());
		for(final RealElement realElement : _realElements.keySet()) {
			final ValueProvider valueProvider = _realElements.get(realElement);
			vc.put(realElement, valueProvider.getValue());
		}
		_realElementValues.push(vc);
	}

	/**
	 * Bereitet die Basisdaten auf und erstellt Ausgangsdatensätze. Wertet alle im Puffer liegenden Wert aus.
	 *
	 * @param done Das Sammeln der Daten wurde abgeschlossen.
	 *
	 * @throws InterruptedException Datenaufbereitung wurde abgebrochen
	 * @throws FailureException     Fehler beim Aufbereiten der Daten.
	 */
	public void createIntermediateDataSet(final boolean done) throws InterruptedException, FailureException {
		/*
		 * Alle vorliegenden Datensätze abarbeiten. So wird sichergestellt, dass keine Datensätze verloren gehen.
		 * Solange noch nicht alle Datensätze eingegangen sind, müssen mindestens zwei Datensätze vorliegen. Ansonsten
		 * kann der Wert des Zeitdauer-Attributs nicht bestimmt werden.
		 */
		while((_realElementValues.size() >= 2) || (done && !_realElementValues.isEmpty())) {
			final ValueContainer vc = _realElementValues.peek();

			// überprüfen ob ein leerer Datensatz eingefügt werden muss.
			if(vc.insertEmpty() && _processingBuffer.isListAggregation()) {
				// JH: Leerer Datensatz macht wenig Sinn, da er nur unter sehr speziellen Bedingungen erzeugt wurde
				// und kaum von einem normalen ProtocolResultData unterscheidbar war.
				// Flag in Datainformation zu Datenlücke reicht für den Client aus.
				_debug.finest("Leerer Datensatz wäre eingefügt worden", vc);

//				_debug.finest("Füge leeren Datensatz ein.");
//				storeEmptyDataSet(vc.getWinnerTimeStamp());
//				_lastBaseData = Collections.emptyList();
			}

			// TODO MH Hier Zeitdauer Attribut berechnen...
			final ValueContainer next = _realElementValues.next();
			/**
			 * next ist null beim allerletzten Datensatz. Deshalb wird der Datensatz als letzter Datensatz markiert.
			 * Dies ist notwenidig um sicherzustellen, dass das Zeitdauer Attribut gesetzt wird. Normalerweise erledigt
			 * der ProcessingBuffer diese Aufgabe. Sollten jedoch sehr viele Datensätze für den Onlinebuffer eingehen,
			 * die nicht alle vor dem Ende des letzten Zeitintervalls verarbeitet werden, ist es möglich, dass<br>
			 * <br>
			 * 1. Genau derjenige Datensatz als letzter Datensatz markiert wird, der bearbeitet wird, als das
			 * Intervallende eingetreten ist.<br>
			 * 2. Weitere Datensätze verarbeitet werden, die zwar vor dem Intervallende eingetroffen sind, jedoch nicht
			 * schnell genug verarbeitet werden konnten.
			 */
			if(next == null && !vc.isLastOfInterval()) vc.setUnknownDuration();
			vc.calcDuration(next);

			// filtern:
			final boolean[] constraints = applyConstraint(vc);

			boolean shouldSendDataSet = false;

			for(final boolean constraint : constraints) {
				shouldSendDataSet |= constraint;
			}

			if(shouldSendDataSet) {
				// Werte in Ausgangsdatensatz eintragen
				final List<BaseDataSet> baseData = calcBaseDataForAllAttributes(vc, constraints);

				// Prüfen ob Werte gleich
				final boolean dataEqual = BaseDataSet.compareBaseData(baseData, _lastBaseData);
				/*
				 * Falls es sich um den letzen Datensatz eines Intervalls gehandelt hat, muss der Intervallzähler
				 * aktualisiert werden.
				 */
				final int interval = _currentInterval;
				if(vc.isLastOfInterval()) _currentInterval += vc.getNextInterval();

				/*
				 * für diesen Datensatz gilt noch das zuletzt verwendete Intervall.
				 */
				_currentStatus = _status.getStatus(baseData, interval);
				ProtocolType protocolType = _processingBuffer.getProtocolType();

				if(protocolType == ProtocolType.StatusProtocol && _noChangeMarkerPerCell){
					// Falls Zustandsprotokoll und keine-Änderungen-Informationen pro Zelle gespeichert werden sollen, hier die
					// Zellen auf NoChange setzen, deren Daten identisch zum vorherigen Datensatz sind.
					List<BaseDataSet> diffDataSet = BaseDataSet.createDiffDataSet(baseData, _lastBaseData);

					if(_processingBuffer.applyAggregations(baseData)) {
						// Aggregationen mit dem originalen Datensatz durchführen

						// Aber aus dem Differenz-Datensatz den IntermediateDataSet bauen
						final IntermediateDataSet ids = _processingBuffer.applyPostFilter(
								diffDataSet, _currentStatus, vc.getWinnerTimeStamp(), interval + _startPeriodIndex
						);
						if(ids != null) {
							// Ausgangsdatensatz ist fertig
							_processingBuffer.storeDataSet(ids);
							_lastBaseData = baseData;
						}
					}
				}
				else if(!dataEqual || protocolType == ProtocolType.EventProtocol)
				{
					// Werte sind unterschiedlich oder Ereignisprotokoll -> Datensatz übernehmen

					if(_processingBuffer.applyAggregations(baseData)) {
						// Nachfilter:
						final IntermediateDataSet ids = _processingBuffer.applyPostFilter(
								baseData, _currentStatus, vc.getWinnerTimeStamp(), interval + _startPeriodIndex
						);
						if(ids != null) {
							// Ausgangsdatensatz ist fertig
							_processingBuffer.storeDataSet(ids);
							_lastBaseData = baseData;
						}
					}
				}
				else if(protocolType == ProtocolType.StatusProtocol && !_merge)
				{
					// Zustandsprotokoll, nicht verschmelzen & Werte sind gleich
				    // -> Quittung senden, dass sich nix geändert hat

					// Werte trotzdem in die Aggregation aufnehmen
					_processingBuffer.applyAggregations(baseData);

					// Hier wird ein Dummy-Datensatz erstellt, der später vor dem Versand durch einen ProtocolNoChanges-Datensatz ersetzt wird.
					// Siehe Processing-Klasse sendProtocolResultData()
					_processingBuffer.storeDataSet(
							new IntermediateDataSet(
									new BaseDataSet[0],
									IntermediateDataSet.NO_CHANGE,
									ApplyAggregation.LISTE,
									vc.getWinnerTimeStamp(),
									interval + _startPeriodIndex
							)
					);
				}
			}
			else {
				_debug.finest("Datensatz wurde ausgefiltert:", vc);
			}
			// Ggf. Intervalldaten sammeln.
			_processingBuffer.storeAggregatedData(_currentStatus);

			/*
			 * Element erst jetzt aus Puffer nehmen Sonst kann es passieren, dass der realElementValues vollständig
			 * geleert wird, und buffer ebenfalls leer ist. Wird genau dann isDone(), liefert dieses true zurück, obwohl
			 * noch Datensätze ausstehen.
			 */
			_realElementValues.pop();
		}
	}

	/**
	 * Erzeugt Ausgangsdatensatz. Berechnet dafür den Wert der temporären Attribute.
	 *
	 * @param valueContainer Reale Attribute und ihre Werte.
	 * @param constraints Filtereinschränkungen, siehe {@link #isNotFiltered(int, boolean[])}
	 * @return Liste mit Datensätzen für diese Zeile.
	 */
	private List<BaseDataSet> calcBaseDataForAllAttributes(final ValueContainer valueContainer, final boolean[] constraints) {
		final List<BaseDataSet> list = new ArrayList<BaseDataSet>(_attributes.size());
		for(final Elements element : _attributes) {
			if(element instanceof RealElement)
			{
				// Attributgruppe oder Reales Attribut
				final RealElement re = (RealElement)element;
				final int index = re.getIndex();
				if(isNotFiltered(index, constraints)) {
					list.add(valueContainer.getValues().get(re));
				}else{
					list.add(createEmptyBaseDataSet());
				}
			}
			else
			{
				// Temporäres Attribut
				final TempAttributeDescription tad = (TempAttributeDescription)element;
				final int index = tad.getIndex();
				if(isNotFiltered(index, constraints)) {
					list.add(_tempElements.get(tad).getValue(valueContainer.getValues(), _tempElements));
				}else{
					list.add(createEmptyBaseDataSet());
				}
			}
		}
		return list;
	}

	/** Stellt fest, ob für die aktuelle Aliasdefinition ein Filter gesetzt ist
	 *
	 * @param rowIndex Der Index des Aliases. -1 für Spalten in denen kein Alias verwendet wird.
	 * @param constraints Für welche Aliase die Bedingung zutrifft. Muss mindestens ein Element haben.
	 * @return true dann, und nur dann, wenn <ul>
	 * <li>constraints die Länge 1 hat (constraints[0] muss dann true sein, sonst würde der ganze Datensatz nicht gesendet werden)</li>
	 * <li>constraints die Länge > 1 hat und rowIndex == -1 ist. Das heißt, das Aliase verwendet werden, aber die aktuelle Spalte keine verwendet und somit die Filterung nach Aliasen keinen Sinn ergibt.</li>
	 * <li>ansonsten wenn constraints[rowindex] true ist. Das heißt wenn der aktuelle Alias ungefiltert ist.</li>
	 * </ul>
	 */
	private boolean isNotFiltered(final int rowIndex, final boolean[] constraints) {
		return constraints.length == 1 || rowIndex == -1 || constraints[rowIndex];
	}

	/**
	 * Wendet den Filterausdruck an.
	 *
	 * @param valueContainer Datensatz, auf den der Filter angewendet werden soll.
	 *
	 * @return Array, das angibt, für welche Aliase die Bedingung erfüllt ist. Diese Spalten werden an den Auftraggeber gesendet. Falls keine Aliase verwendet werden hat das Array ein Element und gilt für die ganze Spalte.
	 *
	 * @throws FailureException Ergebnistyp des Ausdrucks ist nicht {@link de.bsvrz.pua.prot.util.ExpressionResult.ResultType#BOOL}
	 */
	private boolean[] applyConstraint(final ValueContainer valueContainer) throws FailureException {
		if(_constraint != null) {
			final ExpressionTree expression = _constraint.first;
			final int numColumns = expression.getNumColumns();
			final boolean[] resultArray = new boolean[numColumns];
			for(int i = 0; i < numColumns; i++) {
				ExpressionResultAndState expressionResultAndState = expression.evaluateEx(valueContainer.getValues(), _tempElements, i);
				final ExpressionResult expressionResult = expressionResultAndState.getResult();
				PuADataState state = expressionResultAndState.getState();
				if(expressionResult.hasValue()) {
					if(expressionResult.getType() != ResultType.BOOL)
					{
						// falls Ergebnis NONE oder ERROR: kein
						// Fehler (Wert konnte einfach nicht
						// berechnet werden)
						throw new FailureException(
								ProcessingErrorMessage.INVALID_RESULT_TYPE + "bool <-> " + expressionResult.getType() + ", Skript: " + getScriptName(), LoggerException.WARNING
						);
					}
					else {
						resultArray[i] = expressionResult.getBool();
					}
				}
				else {
					// Wenn der Filterausdruck nicht ausgewertet werden konnte, einen Fehler generieren
					if(expressionResult.getType() == ResultType.ERROR) {
						throw new FailureException(
								ProcessingErrorMessage.INVALID_RESULT_TYPE + "bool <-> " + expressionResult.getType() + ": " + expressionResult.getString() + ", Skript: " + getScriptName(),
								LoggerException.WARNING
						);
					}
					else {
						// Datensatz wird durchgelassen
						resultArray[i] = true;
					}
				}
			}
			return resultArray;
		}
		else {
			return new boolean[] {true};
		}
	}

	private BaseDataSet createEmptyBaseDataSet() {
		final DataInformation di = new DataInformation(_processingBuffer.getProtocolType());
		di.dataGap = DataInformation.NO_DATA;
		return new BaseDataSet(RV_FILTERED_DATASET, PuADataState.EMPTY, di, BaseDataSet.INVALID_VALUE, BaseDataSet.INVALID_VALUE, BaseDataSet.INVALID_VALUE, BaseDataSet.INVALID_VALUE, BaseDataSet.INVALID_VALUE, false);
	}

	/**
	 * Zeigt an dass keine weiteren Daten mehr vorliegen.
	 *
	 * @return <code>true</code> falls keine weiteren Daten mehr vorliegen.
	 */
	public boolean isDone() {
		return _realElementValues.isEmpty();
	}

	/**
	 * Markiert den zuletzt eingefügten Ausgangsdatensatz als den letzten des Intervalls.
	 *
	 * @param endOfInterval Endzeitpunkt des Intervalls, das gerade beendet wurde.
	 *
	 * @see ValueContainer#setLastOfInterval(long)
	 */
	public void nextInterval(final long endOfInterval) {
		this._endOfInterval = endOfInterval;
		if(_realElementValues.getLast() != null) {
			_realElementValues.getLast().setLastOfInterval(_endOfInterval);
		}
		else {
			_currentInterval++; // falls noch kein Datensatz angelegt wurde.
		}
	}

	/**
	 * Liefert den momentanen Status zurück.
	 *
	 * @return Status zwischen 0 und 100.
	 */
	public byte getStatus() {
		return _currentStatus;
	}

	/**
	 * Verwirft den zuletzt eingefügten Datensatz
	 * @throws InterruptedException
	 */
	public void throwAway() throws InterruptedException {
		_realElementValues.pop();
	}

	/**
	 * Liefert die Bezeichnung des Skripts zur Verwendung in Fehlerausgaben.
	 * @return Bezeichnung des Skripts
	 */
	private String getScriptName() {
		return _processingBuffer.getScriptName();
	}

}
