/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util.attributes;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeListDefinition;
import de.bsvrz.dav.daf.main.config.AttributeSet;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.impl.subscription.SenderSubscription;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Speichert ein Attribut mit dazugehörigem Index und eventuellen Unterattributen 
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public class AttributeIdentifier implements Cloneable{

	private final Attribute _attribute;

	private final AttributeIdentifier _child;

	/**
	 * Index. Falls kein Index benutzt wird, Integer.MIN_VALUE
	 */
	private final int _index;

	private static final Pattern PATTERN = Pattern.compile("^\\s*(\\S*)\\s*\\[\\s*(\\d*)\\s*\\]\\s*$");

	public AttributeIdentifier(final Attribute attribute) {
		if(attribute == null) throw new IllegalArgumentException("attribute darf nicht null sein.");
		_attribute = attribute;
		_index = Integer.MIN_VALUE;
		_child = null;
	}

	public AttributeIdentifier(final Attribute attribute, final int index) {
		if(attribute == null) throw new IllegalArgumentException("attribute darf nicht null sein.");
		_attribute = attribute;
		if(index < 0) throw new IllegalArgumentException("index muss 0 oder größer sein.");
		_index = index;
		_child = null;
	}

	public AttributeIdentifier(final Attribute attribute, final AttributeIdentifier child) {
		if(attribute == null) throw new IllegalArgumentException("attribute darf nicht null sein.");
		_attribute = attribute;
		_index = Integer.MIN_VALUE;
		_child = child;
	}

	/**
	 * Erzeugt einen AttributeIdentifier aus einem String der Form
	 * Attribut([Index])?.(Attribut([Index])?)*
	 * also z.B.
	 * qKfz.Wert[3]
	 * oder
	 * A.B.C.D
	 * oder
	 * A[17].B.C[11].D
	 * usw.
	 * @param description
	 */
	public AttributeIdentifier(final String description, final AttributeSet set) throws SemanticErrorException {

		if(set == null){
			throw new SemanticErrorException(InterpreterErrorMessage.ATTRIBUTEGROUP_MISSING);
		}

		final int delimeterPos = description.indexOf('.');
		final String part; // Der Teilstring bis zum ersten Punkt.
		if(delimeterPos == -1) {
			part = description;
		}
		else{
			part = description.substring(0, delimeterPos);
		}
		final Matcher matcher = PATTERN.matcher(part);
		final String attributeName;
		if(matcher.find()){
			// Es befinden sich eckige Klammern am Ende
			final String indexString = matcher.group(2);
			try{
				_index = Integer.parseInt(indexString);
			}
			catch(NumberFormatException ignored){
				throw new SemanticErrorException(InterpreterErrorMessage.INVALID_INDEX + matcher.group(1));
			}
			attributeName = matcher.group(1);
		}
		else{
			_index = Integer.MIN_VALUE;
			attributeName = part;
		}

		_attribute = set.getAttribute(attributeName);

		if(_attribute == null){
			throw new SemanticErrorException(InterpreterErrorMessage.ATTRIBUTE_NOT_FOUND + attributeName + " (Attributgruppe: " + set + ")");
		}

		if(delimeterPos == -1) {
			_child = null;
		}
		else{
			if(_attribute.getAttributeType() instanceof AttributeSet) {
				final AttributeSet attributeSet = (AttributeSet)_attribute.getAttributeType();
				_child = new AttributeIdentifier(description.substring(delimeterPos+1), attributeSet);
			}
			else{
				throw new SemanticErrorException(InterpreterErrorMessage.ATTRIBUTE_NOT_LIST + _attribute.getNameOrPidOrId());
			}
		}

		// Prüfen, ob eckige Klammern gültig sind. Das ist nur der Fall, wenn es sich um ein Array-Attribut handelt.
		if(_index != Integer.MIN_VALUE){
			if(!_attribute.isArray()){
				throw new SemanticErrorException(InterpreterErrorMessage.ATTRIBUTE_NOT_ARRAY + _attribute.getNameOrPidOrId());
			}
		}
		else {
			if(_attribute.isArray() && _child != null){
				// Es wurde auf untergeordnete Attribute verwiesen, aber vergessen ein Index anzugeben
				throw new SemanticErrorException(InterpreterErrorMessage.ATTRIBUTE_INDEX_MISSING + _attribute.getNameOrPidOrId());
			}
		}
	}

	public String getName(){
		return _attribute.getName();
	}

	public Attribute getAttribute() {
		return _attribute;
	}

	public int getIndex() {
		return _index;
	}

	public boolean hasIndex() {
		return _index != Integer.MIN_VALUE;
	}

	public AttributeIdentifier getChild() {
		return _child;
	}

	public boolean hasChild() {
		return getChild() != null;
	}

	@Override
	public String toString() {
		return getName() + (hasIndex() ? "[" + getIndex() + "]" : "") + (hasChild()? "." + getChild().toString() : "");
	}

	@Override
	public int hashCode() {
		if(hasChild()) return getAttribute().hashCode()  ^ getIndex() ^ getChild().hashCode() * 37;
		return getAttribute().hashCode() ^ getIndex();
	}

	@Override
	public boolean equals(final Object obj) {
		if(obj instanceof AttributeIdentifier) {
			final AttributeIdentifier other = (AttributeIdentifier)obj;
			if(other.getAttribute().equals(getAttribute()) && other.getIndex() == getIndex()){
				if(hasChild() && other.hasChild()){
					return getChild().equals(other.getChild());
				}
				return !(hasChild() || other.hasChild());
			}
		}
		return false;
	}

	public boolean isAtomic() {
		// Wenn ein Kindelement existiert, dieses fragen
		if(hasChild()) return getChild().isAtomic();

		// Wenn das Attribut ein Array ist, es is nur atomar, wenn kein Index angegeben ist
		if(_attribute.isArray() && !hasIndex()) return false;

		// Ansonsten ist es atomar, wenn es kein Listen-Attribut ist
		return !(_attribute.getAttributeType() instanceof AttributeListDefinition);
	}

	public Attribute getLastAttribute() {
		if(hasChild()) return getChild().getLastAttribute();
		return getAttribute();
	}

	/** Findet das durch diese Klasse referenzierte Datenobjekt (bzw. Unterdatum) in einem Datenobjekt wie es von der Konfiguration kommt. */
	public Data getDataElement(final Data data) {
		Data dataChild = data.getItem(_attribute.getName());
		if(hasIndex()) dataChild = dataChild.asArray().getItem(getIndex());
		if(hasChild()) {
			return getChild().getDataElement(dataChild);
		}
		else{
			return dataChild;
		}
	}
}
